/* arc_pwr_ricefading.ps.c */
/* Alternate received power model for radio link Transceiver Pipeline */

/***************************************************************************
 *
 *           Copyright 2001 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * Author:  Ratish J. Punnoose
 * 
 ***************************************************************************/


#include "opnet.h"
#include <math.h>
#include <stdlib.h>

#define ARC_PI 3.14159265358979323846

/***** constants *****/

#define C		3.0E+08			/* speed of light (m/s) */
#define SIXTEEN_PI_SQ	157.91367		/* 16 times pi-squared */

static const char*	PowI_Err_Hdr = "Error in radio power computation pipeline stage (arc_pwr_ricefading):";

static Compcode LoadDataFile(const char *filename);


static	int N;                  /* Num points in table */
static  float fm0;              /* Max doppler freq in table */
static	float fs;               /* Sampling rate */
static	float dt;               /* Sampling period = 1/fs */
	
static	float *data1;           /* Data values for inphase and quad phase */
static	float *data2;


/***** pipeline procedure *****/

#if defined (__cplusplus)
extern "C"
#endif
void
arc_pwr_ricefading (Packet * pkptr)
{
	double		prop_distance, rcvd_power, path_loss;
	double		tx_power, tx_base_freq, tx_bandwidth, tx_center_freq;
	double		lambda, rx_ant_gain, tx_ant_gain;
	Boolean		sig_lock;
	Objid		rx_ch_obid;
	double		in_band_tx_power, band_max, band_min;
	double		rx_base_freq, rx_bandwidth;
	Objid 		rx_nodeid, tx_nodeid;
	char 		msgbuf[512];
	double 		maxvel;
	double 		riceanK;
	int 		use_two_ray = 0;
	int 		table_offset = 0;

	static int rice_initialized = 0;
	
	/** Compute the average power in Watts of the		**/
	/** signal associated with a transmitted packet.	**/
	FIN (arc_pwr_ricefading (pkptr));


	rx_nodeid = op_td_get_int(pkptr, OPC_TDA_RA_RX_OBJID);
	tx_nodeid = op_td_get_int(pkptr, OPC_TDA_RA_TX_OBJID);


	/* We have to make sure that the initialization is done once
	   even when there are multiple threads */

	if (!rice_initialized)
	{
		char ricefilename[4096];


		if (op_ima_obj_attr_get( rx_nodeid, "RiceDataFile",
					 ricefilename) ==
		    OPC_COMPCODE_FAILURE) 
		{
		    op_sim_end (PowI_Err_Hdr, "No Ricean Data File",
				OPC_NIL, OPC_NIL);
		}


		//sprintf(ricefilename, "%s.gdf", ricefilename);
		
		sprintf(msgbuf, "V 1.1 MaxVelocity = %f  RiceanK = %f",
			maxvel, riceanK);
		op_sim_message(msgbuf, ricefilename);
		

		if( LoadDataFile(ricefilename) ==
		    OPC_COMPCODE_FAILURE) 
		{
		    op_sim_end(PowI_Err_Hdr, "Ricean Data Loading Failed",
				OPC_NIL, OPC_NIL);
		}
		rice_initialized = 1;
	}


	if (op_ima_obj_attr_get( rx_nodeid, "MaxVelocity",
				 &maxvel) ==
	    OPC_COMPCODE_FAILURE) 
	{
		op_sim_end (PowI_Err_Hdr, "No MaxVelocity attr",
			    OPC_NIL, OPC_NIL);
	}

	if (op_ima_obj_attr_get( rx_nodeid, "Ricean KFactor",
				 &riceanK) ==
	    OPC_COMPCODE_FAILURE) 
	{
		op_sim_end (PowI_Err_Hdr, "No Ricean KFactor",
			    OPC_NIL, OPC_NIL);
	}

	if (op_ima_obj_attr_get( rx_nodeid, "UseTwoRay",
				 &use_two_ray) ==
	    OPC_COMPCODE_FAILURE) 
	{
		op_sim_end (PowI_Err_Hdr, "Missing Parameter - UseTwoRay",
			    OPC_NIL, OPC_NIL);
	}

	if (op_ima_obj_attr_get( rx_nodeid, "TableOffset",
				 &table_offset) ==
	    OPC_COMPCODE_FAILURE) 
	{
		sprintf(msgbuf, "No TableOffset Param => Using default 0");
		op_sim_message(msgbuf, 0);
		table_offset = 0;
	}


	/* If the incoming packet is 'valid', it may cause the receiver to */
	/* lock onto it. However, if the receiving node is disabled, then  */
	/* the channel match should be set to noise. */

	if (op_td_get_int (pkptr, OPC_TDA_RA_MATCH_STATUS) ==
	    OPC_TDA_RA_MATCH_VALID) 
	{
		if (op_td_is_set (pkptr, OPC_TDA_RA_ND_FAIL))
		{
			/* The receiving node is disabled.  Change	*/
			/* the channel match status to noise.		*/
			op_td_set_int (pkptr, OPC_TDA_RA_MATCH_STATUS,
				       OPC_TDA_RA_MATCH_NOISE); 
		}
		else
		{
			/* The receiving node is enabled.  Get	*/
			/* the address of the receiver channel.	*/
			rx_ch_obid = op_td_get_int (pkptr,
						    OPC_TDA_RA_RX_CH_OBJID); 
	
			/* If the receiver channel is already locked,
			   the packet will now be considered to be noise.
			   This prevents simultaneous reception of multiple
			   valid packets on any given radio channel.*/
			if (op_ima_obj_attr_get (rx_ch_obid, 
						 "signal lock", &sig_lock) 
			    == OPC_COMPCODE_FAILURE)
			{
				op_sim_end (PowI_Err_Hdr,
					    "Unable to get signal lock flag from channel attribute.",
					    OPC_NIL, OPC_NIL);
			}		
			if (sig_lock)
				op_td_set_int (pkptr,
					       OPC_TDA_RA_MATCH_STATUS, 
					       OPC_TDA_RA_MATCH_NOISE);
			else
			{
				/* Otherwise, the receiver channel will become*/
				/* locked until the packet reception ends. */
				sig_lock = OPC_BOOLINT_ENABLED;
				if (op_ima_obj_attr_set (rx_ch_obid, 
							 "signal lock", 
							 sig_lock)
				    == OPC_COMPCODE_FAILURE)
				{
					op_sim_end (PowI_Err_Hdr,
						    "Unable to set receiver channel attribute (signal lock).",
						    OPC_NIL, OPC_NIL);
				}		
			}
		}
	}



	/* Get power allotted to transmitter channel. */
	tx_power = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_POWER);

	/* Get transmission frequency in Hz. */
	tx_base_freq = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_FREQ);
	tx_bandwidth = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_BW);
	tx_center_freq = tx_base_freq + (tx_bandwidth / 2.0);


	/* When using TMM, the TDA OPC_TDA_RA_RCVD_POWER will already	*/
	/* have a raw value for the path loss. */
	if (op_td_is_set (pkptr, OPC_TDA_RA_RCVD_POWER)) {
		path_loss = op_td_get_dbl (pkptr,
					   OPC_TDA_RA_RCVD_POWER);
		goto done_rawpathloss;
	}


	if(  !(prop_distance > 0.0)) {
		path_loss = 1.0;
		goto done_rawpathloss;
	}


	/* Caclculate wavelength (in meters). */
	lambda = C / tx_center_freq;
	/* Get distance between transmitter and receiver (in meters). */
	prop_distance = op_td_get_dbl (pkptr, OPC_TDA_RA_START_DIST);


	if (!use_two_ray) {
	    path_loss = (lambda * lambda) / (SIXTEEN_PI_SQ *
					     prop_distance *
					     prop_distance);  
	}
	else  {
  	        /* Base Two ray model */
       	        double crossover_dist; 
		double ht, hr, dummy1, dummy2, dummy3, dummy4, dummy5,
			dummy;
	    
		
		if( op_ima_obj_pos_get(op_topo_parent(rx_nodeid),
				       &dummy, &dummy, &hr, 
				       &dummy, &dummy, &dummy) ==
		    OPC_COMPCODE_FAILURE) 
			op_sim_end (PowI_Err_Hdr, "Can't retrieve rxnode",
				    OPC_NIL, OPC_NIL);

		if( op_ima_obj_pos_get(op_topo_parent(tx_nodeid),
				       &dummy, &dummy, &ht, 
				       &dummy, &dummy, &dummy) ==
		    OPC_COMPCODE_FAILURE) 
			op_sim_end (PowI_Err_Hdr, "Can't retrieve txnode",
				    OPC_NIL, OPC_NIL);

		/* sprintf(msgbuf, "ht = %g , hr = %g ", ht, hr);
		   op_sim_message(msgbuf, 0);
		*/


		crossover_dist = (4.0 * ARC_PI * ht * hr) / lambda;

		/* Friis Equation */
		if( prop_distance <= crossover_dist || ht <= 0.0 || hr <= 0.0 ) {
			path_loss = (lambda * lambda) / 
				(SIXTEEN_PI_SQ * 
				 prop_distance * prop_distance);
			goto done_rawpathloss;
		}

		/* If we get to here, we are doing a two-ray model */
		path_loss = hr * hr * ht * ht / ( prop_distance *
						  prop_distance *
						  prop_distance *
						  prop_distance);

	}


 done_rawpathloss:
	/* Now working on Ricean path loss */
	{

		double time_index, fm;
		double envelope_fac, tmp_x1, tmp_x2, x1_interp, x2_interp;

		fm = maxvel / lambda;

		time_index = ( op_sim_time() *
				        fs *   fm / fm0) ;

		time_index = time_index -
			((double)N)*floor(time_index/((double)N));

		time_index += table_offset;

		{
			/* Do envelope interpolation using Legendre
			   polynomials */
			double X0, X1, X2, X3;
			int ind0, ind1, ind2, ind3;
			
			ind1 = (int)(floor(time_index));
			ind1 = ind1 % N;
			ind0 = (ind1-1+N) % N;
			ind2 = (ind1+1) % N;
			ind3 = (ind1+2) % N;
			
			X1 = time_index - ind1;
			X0 = X1+1.0;
			X2 = X1-1.0;
			X3 = X1-2.0;

			x1_interp = data1[ind0]*X1*X2*X3/(-6.0) +
				data1[ind1]*X0*X2*X3*(0.5) +
				data1[ind2]*X0*X1*X3*(-0.5) +
				data1[ind3]*X0*X1*X2/6.0;
				
			x2_interp = data2[ind0]*X1*X2*X3/(-6.0) +
				data2[ind1]*X0*X2*X3*(0.5) +
				data2[ind2]*X0*X1*X3*(-0.5) +
				data2[ind3]*X0*X1*X2/6.0;
			
		}

		/* Find the envelope multiplicative factor */
		tmp_x1 = x1_interp + sqrt(2.0 * riceanK);
		tmp_x2 = x2_interp;

		envelope_fac = (tmp_x1*tmp_x1 + tmp_x2*tmp_x2) / 
			(2.0 * (riceanK+1.0)); 

		//Pr_Rice = 10.0 * log10(envelope_fac);
		// Pr_Rice = envelope_fac;

		path_loss = path_loss * envelope_fac;



	}

 done_ricepathloss:


	/* Determine the receiver bandwidth and base frequency. */
	rx_base_freq = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_FREQ);
	rx_bandwidth = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_BW);

	/* Use these values to determine the band overlap with the transmitter.	*/
	/* Note that if there were no overlap at all, the packet would already	*/
	/* have been filtered by the channel match stage.						*/

	/* The base of the overlap band is the highest base frequency. */
	if (rx_base_freq > tx_base_freq) 
		band_min = rx_base_freq;
	else
		band_min = tx_base_freq;
	
	/* The top of the overlap band is the lowest end frequency. */
	if (rx_base_freq + rx_bandwidth > tx_base_freq + tx_bandwidth)
		band_max = tx_base_freq + tx_bandwidth;
	else
		band_max = rx_base_freq + rx_bandwidth;

	/* Compute the amount of in-band transmitter power. */
	in_band_tx_power = tx_power * (band_max - band_min) / tx_bandwidth;

	/* Get antenna gains (raw form, not in dB). */
	tx_ant_gain = pow (10.0, op_td_get_dbl (pkptr, OPC_TDA_RA_TX_GAIN) / 10.0);
	rx_ant_gain = pow (10.0, op_td_get_dbl (pkptr, OPC_TDA_RA_RX_GAIN) / 10.0);

	/* Calculate received power level. */
	rcvd_power = in_band_tx_power * tx_ant_gain * path_loss * rx_ant_gain;

	

	/* Assign the received power level (in Watts) */
	/* to the packet transmission data attribute. */
	op_td_set_dbl (pkptr, OPC_TDA_RA_RCVD_POWER, rcvd_power);



	/* sprintf(msgbuf, "time = %3.9f , rcvd_power = %g",
	   op_sim_time(), rcvd_power);  
	   op_sim_message(msgbuf, 0);
	*/


	FOUT;
}




#define BUF_SZ 4096
static Compcode LoadDataFile(const char *filename)
{
	char buf[BUF_SZ];
	char arg1[BUF_SZ];
	char arg2[BUF_SZ];

	char *ret;
	FILE *fdstream;
	float *tmp1, *tmp2;
	int rc = OPC_COMPCODE_FAILURE, ret_val;
	int k = 0;


	List* rice_line_ptr;
	int   num_lines;
	int   linenum;

	FIN (LoadDataFile(filename)) ;

	op_sim_message("Opening file now\n", 0);


	rice_line_ptr = op_prg_gdf_read( filename );
	
	if( rice_line_ptr == OPC_NIL) {
	    op_sim_message("Couldn't open file", 0);
	    goto quit;
	}		

	num_lines = op_prg_list_size( rice_line_ptr );
	linenum = 0;

	while(1) {
		if( linenum >= num_lines)
			goto quit;

		sscanf((char*) op_prg_list_access(rice_line_ptr,
						  linenum), 
		       "%s %s", arg1, arg2); 
		linenum++;
		if(!strcmp(arg1, "DATA")) {
			break;
		}

		if(!strcmp(arg1, "fm")) {
			fm0 = atof(arg2);
		} else if(!strcmp(arg1, "fs")) {
			fs = atof(arg2);
		} else if(!strcmp(arg1, "N")) {
			N = atoi(arg2);
		} 


	}
	
	data1 = (float *) malloc(N * sizeof(float));
	data2 = (float *) malloc(N * sizeof(float));

	tmp1 = data1;
	tmp2 = data2;

	for(k=0; k<N; k++) {
		if( linenum >= num_lines)
			goto quit;

		ret_val = sscanf((char*) op_prg_list_access(rice_line_ptr,
						  linenum), 
		       "%f %f", tmp1++, tmp2++); 
		linenum++;
		
		if(ret_val != 2) {
			op_sim_message("File Format error", 0);
			goto quit;
		}
	}

	rc = OPC_COMPCODE_SUCCESS;
			
 quit:
	//printf("%d data points read\n", k);
	//printf("fm0 = %f fm = %f  fs = %f\n", fm0, fm, fs);



	FRET ( rc );

}

